"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.CrysteamPortal = exports.Methods = void 0;
const fs_1 = __importDefault(require("fs"));
const jsonwebtoken_1 = __importDefault(require("jsonwebtoken"));
const crypto_1 = __importDefault(require("crypto"));
const axios_1 = __importDefault(require("axios"));
const node_cache_1 = __importDefault(require("node-cache"));
const path_to_regexp_1 = require("path-to-regexp");
var Methods;
(function (Methods) {
    Methods["POST"] = "POST";
    Methods["PUT"] = "PUT";
    Methods["DELETE"] = "DELETE";
    Methods["PATCH"] = "PATCH";
    Methods["GET"] = "GET";
})(Methods || (exports.Methods = Methods = {}));
class CrysteamPortal {
    constructor(options) {
        this._tokenCache = new node_cache_1.default({ stdTTL: 1800 });
        this._portalHost = options.portalHost || "https://portal.crysteam.ru/api/v2";
        this._metaFilePath = options.metaFilePath || "./data/.portal";
        this._hostname = options.hostname;
        this._secretKey = Buffer.from(`${options.clientId}:${options.clientSecret}`).toString("base64");
        this._axios = axios_1.default.create({
            baseURL: this._portalHost,
            headers: { authorization: `Bearer ${this._secretKey}` }
        });
        this._initialize(options);
    }
    _initialize(options) {
        return __awaiter(this, void 0, void 0, function* () {
            if (options.label)
                yield this.registerService(options.label, options.description);
        });
    }
    sign(text) {
        if (!this._serviceMeta)
            throw new Error("[Portal] Meta not loaded");
        const signer = crypto_1.default.createSign("RSA-SHA256");
        signer.update(text);
        signer.end();
        return signer.sign(this._serviceMeta.clientPrivateKey, "base64");
    }
    buildSignedPayload(data = {}) {
        if (!this._serviceMeta)
            throw new Error("[Portal] Meta not loaded");
        const signature = this.sign(JSON.stringify(data));
        return {
            service_id: this._serviceMeta.serviceId,
            signature,
            payload: data,
        };
    }
    generateRsaKeyPair() {
        console.log("[Portal] Generating RSA key pair...");
        return crypto_1.default.generateKeyPairSync("rsa", {
            modulusLength: 2048,
            publicKeyEncoding: { type: "spki", format: "pem" },
            privateKeyEncoding: { type: "pkcs8", format: "pem" },
        });
    }
    saveMetaToFile(meta) {
        console.log(`[Portal] Saving meta to ${this._metaFilePath}`);
        fs_1.default.mkdirSync(require("path").dirname(this._metaFilePath), { recursive: true });
        fs_1.default.writeFileSync(this._metaFilePath, JSON.stringify(meta, null, 2), "utf8");
    }
    loadMetaFromFile() {
        console.log(`[Portal] Loading meta from ${this._metaFilePath}`);
        if (!fs_1.default.existsSync(this._metaFilePath))
            return;
        try {
            return this._serviceMeta = JSON.parse(fs_1.default.readFileSync(this._metaFilePath, "utf8"));
        }
        catch (err) {
            console.error("[Portal] Failed to parse meta:", err);
        }
    }
    getScopes() {
        return __awaiter(this, void 0, void 0, function* () {
            const { data } = yield this._axios.post("/services/scopes", this.buildSignedPayload());
            return data;
        });
    }
    registerService(label, description) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.loadMetaFromFile() || this._serviceMeta)
                return;
            const { privateKey, publicKey } = this.generateRsaKeyPair();
            const { data } = yield this._axios.post("/services/register", {
                label,
                description,
                publicKey,
                hostname: this._hostname,
            });
            const meta = {
                clientPrivateKey: privateKey,
                portalPublicKey: data.publicKey,
                serviceId: data.id,
            };
            this.saveMetaToFile(meta);
            this._serviceMeta = meta;
            return data.id;
        });
    }
    generateAccessToken(serviceFrom, scope) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this._serviceMeta)
                throw new Error("[Portal] Meta not loaded");
            const cacheKey = `${serviceFrom}:${JSON.stringify(scope)}`;
            const cached = this._tokenCache.get(cacheKey);
            if (cached)
                return cached.token;
            const tokenPayload = {
                service_from: serviceFrom,
                service_to: this._serviceMeta.serviceId,
                scope,
            };
            const signature = this.sign(JSON.stringify(tokenPayload));
            console.log("[Portal] Requesting access token...");
            const { data: token } = yield this._axios.post("/services/produce", Object.assign(Object.assign({}, tokenPayload), { signature }));
            this._tokenCache.set(cacheKey, { token });
            return token;
        });
    }
    getServiceScopesFor(domain) {
        return __awaiter(this, void 0, void 0, function* () {
            const scopes = yield this.getScopes();
            const domainScopes = scopes[domain];
            if (!domainScopes)
                throw new Error(`[Portal] No scopes found for domain "${domain}"`);
            return domainScopes;
        });
    }
    isScopeAllowed(arg, permission) {
        if (Array.isArray(arg)) {
            return arg.some(scope => scope.method === permission.method &&
                scope.path === permission.path);
        }
        return this.getServiceScopesFor(arg)
            .then(scopes => scopes.some(scope => scope.method === permission.method &&
            scope.path === permission.path))
            .catch(() => false);
    }
    createHttpClientFor(serviceFrom, scope) {
        const client = axios_1.default.create({ baseURL: serviceFrom });
        client.interceptors.request.use((config) => __awaiter(this, void 0, void 0, function* () {
            const token = yield this.generateAccessToken(serviceFrom, scope);
            config.headers['x-service-token'] = `Bearer ${token}`;
            return config;
        }));
        client.interceptors.response.use(undefined, (error) => __awaiter(this, void 0, void 0, function* () {
            var _a;
            if (((_a = error.response) === null || _a === void 0 ? void 0 : _a.status) === 403) {
                const token = yield this.generateAccessToken(serviceFrom, scope);
                error.config.headers['x-service-token'] = `Bearer ${token}`;
                return axios_1.default.request(error.config);
            }
            return Promise.reject(error);
        }));
        return client;
    }
    verifyApiKey() {
        return (req, res, next) => {
            try {
                if (!this._hostname)
                    throw new Error("Hostname is not configured.");
                const fullPatch = req.baseUrl + req.path;
                console.log(`[Portal][Auth] ${req.method} ${fullPatch}`);
                const serviceToken = req.headers['x-service-token'];
                if (!serviceToken)
                    throw new Error("Missing token.");
                const token = serviceToken.split(" ")[1];
                if (!token)
                    throw new Error("Missing token.");
                const meta = this.loadMetaFromFile();
                if (!meta)
                    throw new Error("Missing meta information.");
                const decoded = jsonwebtoken_1.default.verify(token, meta.portalPublicKey, {
                    algorithms: ["RS256"],
                });
                const isAllowed = decoded.scope.some(({ method, path }) => {
                    if (decoded.service_from !== this._hostname)
                        return false;
                    if (method !== req.method)
                        return false;
                    return (0, path_to_regexp_1.match)(path, { decode: decodeURIComponent })(fullPatch) !== false;
                });
                if (!isAllowed)
                    throw new Error("Access denied.");
                req.user = decoded;
                next();
            }
            catch (err) {
                next(err);
            }
        };
    }
}
exports.CrysteamPortal = CrysteamPortal;
